package com.yexuejc.base.util;

import java.text.ParseException;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.TemporalAdjusters;
import java.util.Date;


/**
 * 新版时间日期出来 工具
 *
 * @author: maxf
 * @date: 2018/3/27 10:44
 */
public class DateTimeUtil {
    public static String DATE_PATTERN = "yyyy-MM-dd";
    public static String TIME_PATTERN = "HH:mm:ss";
    public static String DATE_TIME_PATTERN = "yyyy-MM-dd HH:mm:ss";
    public static String DATE_TIME_MS_PATTERN = "yyyy-MM-dd HH:mm:ss.SSS";

    private DateTimeUtil() {
    }

    /**
     * 获取本年第一天
     *
     * @return
     */
    public static LocalDate getYear4First() {
        return getYear4First(LocalDate.now());
    }

    /**
     * 指定日期 年第一天
     *
     * @param now
     * @return
     */
    public static LocalDate getYear4First(LocalDate now) {
        return now.with(TemporalAdjusters.firstDayOfYear());
    }

    /**
     * 获取本年最后一天
     *
     * @return
     */
    public static LocalDate getYear4Last() {
        return getYear4Last(LocalDate.now());
    }

    /**
     * 指定日期 年最后一天
     *
     * @param now
     * @return
     */
    public static LocalDate getYear4Last(LocalDate now) {
        return now.with(TemporalAdjusters.lastDayOfYear());
    }


    /**
     * 获取本月第一天
     *
     * @return
     */
    public static LocalDate getMonth4First() {
        return getMonth4First(LocalDate.now());
    }

    /**
     * 指定日期 月第一天
     *
     * @param now
     * @return
     */
    public static LocalDate getMonth4First(LocalDate now) {
        return now.with(TemporalAdjusters.firstDayOfMonth());
    }

    /**
     * 获取本月第最后一天
     *
     * @return
     */
    public static LocalDate getMonth4Last() {
        return getMonth4Last(LocalDate.now());
    }

    /**
     * 指定日期 月第最后一天
     *
     * @param now
     * @return
     */
    public static LocalDate getMonth4Last(LocalDate now) {
        return now.with(TemporalAdjusters.lastDayOfMonth());
    }

    /**
     * 获取本周第一天
     *
     * @return {@link LocalDate}
     */
    public static LocalDate getWeek4First() {
        return getWeek4First(LocalDate.now());
    }

    /**
     * 指定日期 周第一天
     *
     * @param date
     * @return
     */
    public static LocalDate getWeek4First(LocalDate date) {
        TemporalAdjuster firstOfWeek = TemporalAdjusters.ofDateAdjuster(localDate ->
                localDate.minusDays(localDate.getDayOfWeek().getValue() - DayOfWeek.MONDAY.getValue()));
        return date.with(firstOfWeek);
    }

    /**
     * 获取本周最后一天
     *
     * @return
     */
    public static LocalDate getWeek4Last() {
        return getWeek4Last(LocalDate.now());
    }

    /**
     * 指定日期 周最后一天
     *
     * @param date
     * @return
     */
    public static LocalDate getWeek4Last(LocalDate date) {
        TemporalAdjuster lastOfWeek = TemporalAdjusters.ofDateAdjuster(localDate ->
                localDate.plusDays(DayOfWeek.SUNDAY.getValue() - localDate.getDayOfWeek().getValue()));
        return date.with(lastOfWeek);
    }

    /**
     * ZonedDateTime 转 Date
     *
     * @param zonedDateTime
     * @return
     */
    public static Date parseDate(ZonedDateTime zonedDateTime) {
        ZoneId zoneId = ZoneId.systemDefault();
        ZonedDateTime zdt = zonedDateTime.withZoneSameInstant(zoneId);
        Date date = Date.from(zdt.toInstant());
        return date;
    }

    /**
     * LocalDate 转 Date
     *
     * @param localDate
     * @return
     */
    public static Date parseDate(LocalDate localDate) {
        ZoneId zone = ZoneId.systemDefault();
        Instant instant = localDate.atStartOfDay().atZone(zone).toInstant();
        return Date.from(instant);
    }

    /**
     * LocalDateTime 转 Date
     *
     * @param localDateTime
     * @return
     */
    public static Date parseDate(LocalDateTime localDateTime) {
        ZoneId zoneId = ZoneId.systemDefault();
        ZonedDateTime zdt = localDateTime.atZone(zoneId);
        return Date.from(zdt.toInstant());
    }

    /**
     * LocalTime 转 Date
     *
     * @param localTime 时间
     * @return 当前日期的指定时间
     */
    public static Date parseDate(LocalTime localTime) {
        LocalDate localDate = LocalDate.now();
        return parseDate(localDate, localTime);
    }

    /**
     * LocalDate + LocalTime 转 Date
     *
     * @param localDate 日期
     * @param localTime 时间
     * @return 指定日期的指定时间
     */
    public static Date parseDate(LocalDate localDate, LocalTime localTime) {
        LocalDateTime localDateTime = LocalDateTime.of(localDate, localTime);
        ZoneId zone = ZoneId.systemDefault();
        Instant instant = localDateTime.atZone(zone).toInstant();
        return Date.from(instant);
    }

    /**
     * Date 转 ZonedDateTime
     *
     * @param date
     * @return
     */
    public static ZonedDateTime parseZonedDateTime(Date date) {
        Instant instant = date.toInstant();
        ZoneId zoneId = ZoneId.systemDefault();
        return instant.atZone(zoneId).withZoneSameInstant(zoneId);
    }

    /**
     * Date 转 LocalDateTime
     *
     * @param date
     * @return
     */
    public static LocalDateTime parseLocalDateTime(Date date) {
        Instant instant = date.toInstant();
        ZoneId zoneId = ZoneId.systemDefault();
        return instant.atZone(zoneId).toLocalDateTime();
    }

    /**
     * Date 转 LocalDate
     *
     * @param date
     * @return
     */
    public static LocalDate parseLocalDate(Date date) {
        Instant instant = date.toInstant();
        ZoneId zoneId = ZoneId.systemDefault();
        return instant.atZone(zoneId).toLocalDate();
    }

    /**
     * Date 转 LocalTime
     *
     * @param date
     * @return
     */
    public static LocalTime parseLocalTime(Date date) {
        Instant instant = date.toInstant();
        ZoneId zone = ZoneId.systemDefault();
        LocalDateTime localDateTime = LocalDateTime.ofInstant(instant, zone);
        return localDateTime.toLocalTime();
    }

    /**
     * Long 转 LocalDateTime
     *
     * @param timestamp 13位（毫秒）
     * @return
     */
    public static LocalDateTime parseLocalDateTime13(long timestamp) {
        return parseLocalDateTime13(timestamp, ZoneId.systemDefault());
    }

    /**
     * Long 转 LocalDateTime
     *
     * @param timestamp 13位（毫秒）
     * @param zoneId    时区
     * @return
     */
    public static LocalDateTime parseLocalDateTime13(long timestamp, ZoneId zoneId) {
        Instant instant = Instant.ofEpochMilli(timestamp);
        return LocalDateTime.ofInstant(instant, zoneId);
    }

    /**
     * Long 转 LocalDateTime
     *
     * @param timestamp 10位（秒）
     * @return
     */
    public static LocalDateTime parseLocalDateTime10(long timestamp) {
        return parseLocalDateTime10(timestamp, ZoneId.systemDefault());
    }

    /**
     * Long 转 LocalDateTime
     *
     * @param timestamp 10位（秒）
     * @param zoneId    时区
     * @return
     */
    public static LocalDateTime parseLocalDateTime10(long timestamp, ZoneId zoneId) {
        Instant instant = Instant.ofEpochMilli(timestamp * 1000);
        return LocalDateTime.ofInstant(instant, zoneId);
    }

    /**
     * UTC（格林威治，标准）时间
     *
     * @param timestamp 13位（毫秒）/ 10位（秒）
     * @return
     */
    public static LocalDateTime parserUTC(Long timestamp) {
        if (String.valueOf(timestamp).length() == 10) {
            timestamp = timestamp * 1000;
        }
        return parseLocalDateTime13(timestamp, ZoneId.of("UTC"));
    }

    /**
     * LocalDateTime 转 Long
     *
     * @param localDateTime
     * @return 13位（毫秒）
     */
    public static long parseLong(LocalDateTime localDateTime) {
        ZoneId zone = ZoneId.systemDefault();
        Instant instant = localDateTime.atZone(zone).toInstant();
        return instant.toEpochMilli();
    }

    /**
     * LocalDate 转 Long
     *
     * @param localDate
     * @return 13位（毫秒）最后3位为0
     */
    public static long parseLong(LocalDate localDate) {
        ZoneId zone = ZoneId.systemDefault();
        Instant instant = localDate.atStartOfDay(zone).toInstant();
        return instant.toEpochMilli();
    }


    /**
     * 格式化时间 <br/>
     * 格式 yyyy-MM-dd HH:mm:ss
     *
     * @param dateTime
     * @return
     */
    public static String format(LocalDate dateTime) {
        return format(dateTime, null);
    }

    /**
     * 格式化时间
     *
     * @param dateTime
     * @param pattern  格式 默认:yyyy-MM-dd
     * @return
     */
    public static String format(LocalDate dateTime, String pattern) {
        if (StrUtil.isEmpty(pattern)) {
            pattern = DATE_PATTERN;
        }
        DateTimeFormatter df = DateTimeFormatter.ofPattern(pattern);
        return df.format(dateTime);
    }

    /**
     * 格式化时间 <br/>
     * 格式 yyyy-MM-dd HH:mm:ss
     *
     * @param dateTime
     * @return
     */
    public static String format(LocalDateTime dateTime) {
        return format(dateTime, null);
    }

    /**
     * 格式化时间
     *
     * @param dateTime
     * @param pattern  格式 默认:yyyy-MM-dd HH:mm:ss
     * @return
     */
    public static String format(LocalDateTime dateTime, String pattern) {
        if (StrUtil.isEmpty(pattern)) {
            pattern = DATE_TIME_PATTERN;
        }
        DateTimeFormatter df = DateTimeFormatter.ofPattern(pattern);
        return df.format(dateTime);
    }

    /**
     * 获取UTC（格林威治，标准）时间
     *
     * @param date
     * @return
     */
    public static LocalDateTime convertUTC(LocalDateTime date) {
        return convertTimezone(date, null, "UTC");
    }

    /**
     * 时区转换(当前时区)
     *
     * @param date   时间
     * @param zoneId 转换时区
     * @return
     */
    public static LocalDateTime convertTimezone(LocalDateTime date, String zoneId) {
        return convertTimezone(date, null, zoneId);
    }


    /**
     * 时区转换
     *
     * @param date          时间
     * @param currentZoneId 当前时间的时区
     * @param targetZoneId  转换时区
     * @return
     */
    public static LocalDateTime convertTimezone(LocalDateTime date, String currentZoneId, String targetZoneId) {
        ZoneId currentZone;
        if (currentZoneId == null) {
            currentZone = ZoneId.systemDefault();
        } else {
            currentZone = ZoneId.of(currentZoneId);
        }
        ZoneId targetZone = ZoneId.of(targetZoneId);
        return date.atZone(currentZone).withZoneSameInstant(targetZone).toLocalDateTime();
    }


    /*public static void main(String[] args) throws ParseException {
        System.out.println(parserUTC(1684140338161L));
        System.out.println(convertUTC(LocalDateTime.now()));
        Date date = DateUtil.str2dateTime("2023-05-15 08:28:05.327");
        LocalDateTime localDateTime = parseLocalDateTime(date);
        System.out.println(date);
        System.out.println(localDateTime);
        LocalDateTime utc = convertTimezone(localDateTime, "UTC", "Asia/Shanghai");
        System.out.println(utc);

        System.out.println(getWeek4First());
        System.out.println(format(getWeek4First(LocalDate.parse("2018-02-10")).atTime(LocalTime.MIN)));
        System.out.println(format(getWeek4Last(LocalDate.parse("2018-02-10")).atTime(LocalTime.MAX)));

        System.out.println(format(getMonth4First().atTime(LocalTime.MIN)));
        System.out.println(format(getMonth4Last().atTime(LocalTime.MAX)));

        System.out.println(format(getYear4First().atTime(LocalTime.MIN)));
        System.out.println(format(getYear4Last().atTime(LocalTime.MAX)));

    }*/


}
