package com.yexuejc.base.util;

import com.fasterxml.jackson.databind.MappingIterator;
import com.fasterxml.jackson.dataformat.csv.CsvMapper;
import com.fasterxml.jackson.dataformat.csv.CsvSchema;
import com.google.common.io.ByteStreams;
import com.google.common.io.CharStreams;
import com.yexuejc.base.annotation.CsvHeader;
import com.yexuejc.base.pojo.ReadFileBean;
import io.jsonwebtoken.lang.Assert;
import org.apache.commons.io.IOUtils;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.io.StringWriter;
import java.math.BigInteger;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;
import java.util.Scanner;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.zip.CRC32;

/**
 * 文件工具类
 *
 * @author maxf:yexue
 * @className FileUtil
 * @description 工具类
 * @time 2017年11月3日 下午3:12:49
 */
public class FileUtil {
    static Logger logger = Logger.getLogger(FileUtil.class.getName());

    private FileUtil() {
    }

    private static final String TYPE_TAR_GZ = ".tar.gz";
    private static final String TYPE_CSV = ".csv";
    private static final String TAR_GZ = "tar.gz";

    /**
     * 获取文件类型：不适合所有
     * <p>
     * 根据文件名称截取.后的文件格式
     * </p>
     *
     * @param fileName
     * @return
     */
    public static String getFileType(String fileName) {
        try {
            if (fileName.lastIndexOf(TYPE_TAR_GZ) > 0) {
                return TAR_GZ;
            }
            return fileName.substring(fileName.lastIndexOf(".") + 1);
        } catch (Exception e) {
            logger.severe("file doesn't exist or is not a file");
        }
        return null;
    }

    /**
     * 判断文件是否存在,不存在就创建一个空的
     *
     * @param file
     */
    public static void judeFileExists(File file) {

        if (file.exists()) {
            logger.severe("file exists");
        } else {
            logger.info("file not exists, create it ...");
            try {
                file.createNewFile();
            } catch (IOException e) {
                logger.severe("file create fail");
                e.printStackTrace();
            }
        }

    }

    /**
     * <pre>
     * 判断文件夹是否存在, 不存在创建【采用mkdirs】
     * 相关解释：
     * 1、File类的createNewFile根据抽象路径创建一个新的空文件，当抽象路径制定的文件存在时，创建失败
     * 2、File类的mkdir方法根据抽象路径创建目录
     * 3、File类的mkdirs方法根据抽象路径创建目录，包括创建必需但不存在的父目录
     * 4、File类的createTempFile方法创建临时文件，可以制定临时文件的文件名前缀、后缀及文件所在的目录，如果不指定目录，则存放在系统的临时文件夹下。
     * 5、除mkdirs方法外，以上方法在创建文件和目录时，必须保证目标文件不存在，而且父目录存在，否则会创建失败
     * </pre>
     *
     * @return 创建成功、失败
     */
    public static boolean judeDirExists(File file) {
        if (file.exists()) {
            if (file.isDirectory()) {
                logger.severe("dir exists");
            } else {
                logger.severe("the same name file exists, can not create dir");
            }
        } else {
            logger.info("dir not exists, create it ...");
            return file.mkdirs();
        }
        return false;
    }

    /**
     * 获取文件sha1
     *
     * @param file
     * @return
     */
    public static String sha1(File file) {
        FileInputStream in = null;
        try {
            in = new FileInputStream(file);
            MessageDigest digest = MessageDigest.getInstance("SHA-1");
            byte[] buffer = new byte[1024 * 1024 * 10];

            int len = 0;
            while ((len = in.read(buffer)) > 0) {
                digest.update(buffer, 0, len);
            }
            String sha1 = new BigInteger(1, digest.digest()).toString(16);
            int length = 40 - sha1.length();
            if (length > 0) {
                for (int i = 0; i < length; i++) {
                    sha1 = "0" + sha1;
                }
            }
            return sha1;
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
            }
        }
        return null;
    }

    /***
     * 计算SHA1码
     *
     * @return String 适用于上G大的文件
     * @throws NoSuchAlgorithmException
     * */
    public static String sha1ByBigFile(File file) {
        MessageDigest messagedigest = null;
        try {
            messagedigest = MessageDigest.getInstance("SHA-1");
            FileInputStream in = new FileInputStream(file);
            FileChannel ch = in.getChannel();
            MappedByteBuffer byteBuffer = ch.map(FileChannel.MapMode.READ_ONLY, 0, file.length());
            messagedigest.update(byteBuffer);
            return StrUtil.toHex(messagedigest.digest());
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        }
        return null;
    }

    /**
     * 文件md5
     *
     * @param file
     * @return
     */
    public static String md5(File file) {
        FileInputStream in = null;
        try {
            in = new FileInputStream(file);
            MessageDigest digest = MessageDigest.getInstance("MD5");
            byte[] buffer = new byte[1024 * 1024 * 10];

            int len = 0;
            while ((len = in.read(buffer)) > 0) {
                digest.update(buffer, 0, len);
            }
            String md5 = new BigInteger(1, digest.digest()).toString(16);
            int length = 32 - md5.length();
            if (length > 0) {
                for (int i = 0; i < length; i++) {
                    md5 = "0" + md5;
                }
            }
            return md5;
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
        } finally {
            try {
                if (in != null) {
                    in.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
            }
        }
        return null;
    }

    /**
     * 对一个文件获取md5值
     *
     * @return md5串
     * @throws NoSuchAlgorithmException
     */
    public static String md5ByBigFile(File file) {

        MessageDigest messagedigest = null;
        try {
            messagedigest = MessageDigest.getInstance("MD5");
            FileInputStream in = new FileInputStream(file);
            FileChannel ch = in.getChannel();
            MappedByteBuffer byteBuffer = ch.map(FileChannel.MapMode.READ_ONLY, 0,
                    file.length());
            messagedigest.update(byteBuffer);
            return StrUtil.toHex(messagedigest.digest());
        } catch (NoSuchAlgorithmException e) {
            logger.severe("system algorithm error.");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
        }
        return null;
    }

    /**
     * 获取文件CRC32码
     *
     * @return String
     */
    public static String crc32(File file) {
        CRC32 crc32 = new CRC32();
        // MessageDigest.get
        FileInputStream fileInputStream = null;
        try {
            fileInputStream = new FileInputStream(file);
            byte[] buffer = new byte[8192];
            int length;
            while ((length = fileInputStream.read(buffer)) != -1) {
                crc32.update(buffer, 0, length);
            }
            return crc32.getValue() + "";
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
            e.printStackTrace();
            return null;
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
            e.printStackTrace();
            return null;
        } finally {
            try {
                if (fileInputStream != null) {
                    fileInputStream.close();
                }
            } catch (IOException e) {
                logger.severe("close FileInputStream IO exception.");
                e.printStackTrace();
            }
        }
    }

    /**
     * 获取文件base64
     *
     * @param file
     * @return
     */
    public static String base64ToStr(File file) {
        try {
            byte[] bytes = Files.readAllBytes(Path.of(file.getPath()));
            return Base64.getEncoder().encodeToString(bytes);
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        }
        return null;
    }

    /**
     * base64转文件
     * <p>
     * <i>
     * 文件转base64请使用 {@link FileUtil#base64ToStr(File)}
     * </i>
     *
     * @param decode   {@link FileUtil#base64ToStr(File)} 的结果
     * @param fileName 保存文件名称（包含路径）
     * @return 返回保存地址
     */
    public static String base64ToFile(String decode, String fileName) {
        return base64ToFile(Base64.getDecoder().decode(decode.getBytes()), fileName);
    }

    /**
     * base64转文件
     * <p>
     * <i>
     * 文件转base64请使用 {@link FileUtil#base64ToStr(File)}
     * </i>
     *
     * @param decode   baseByte
     * @param fileName 文件名称（包含路径）
     * @return 返回保存地址
     */
    public static String base64ToFile(byte[] decode, String fileName) {

        FileOutputStream out = null;
        try {
            out = new FileOutputStream(fileName);
            out.write(decode);
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } finally {
            try {
                if (out != null) {
                    out.close();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        return fileName;
    }

    /**
     * 获取文件大小 ：直接返回大小
     *
     * @param f
     * @return f.length()
     */
    public static long size(File f) {
        if (f.exists() && f.isFile()) {
            return f.length();
        } else {
            logger.info("file doesn't exist or is not a file");
        }
        return 0;
    }

    /**
     * 获取文件大小 : 用流的方式获取
     *
     * @param f
     * @return
     */
    public static long size4Stream(File f) {
        FileChannel fc = null;
        try {
            if (f.exists() && f.isFile()) {
                FileInputStream fis = new FileInputStream(f);
                fc = fis.getChannel();
                return fc.size();
            } else {
                logger.info("file doesn't exist or is not a file");
            }
        } catch (FileNotFoundException e) {
            logger.severe("file doesn't exist or is not a file");
        } catch (IOException e) {
            logger.severe("The operation file is an IO exception.");
        } finally {
            if (null != fc) {
                try {
                    fc.close();
                } catch (IOException e) {
                    logger.severe("close FileInputStream IO exception.");
                }
            }
        }
        return 0;
    }

    /**
     * 字符串（csv格式）转 对象
     *
     * @param data      转换的字符串 如
     *                  <p> ------------ </p>
     *                  <p> id,name,age </p>
     *                  <p> 1,zhangsan,18   </p>
     *                  <p> 2,lisi,20 </p>
     *                  <p> ------------ </p>
     * @param cls       需要转换的对象,含有<b>id,name,age</b>字段
     * @param delimiter 分隔符
     * @param <I>
     * @return
     */
    public static <I> List<I> readCsv(String data, Class<I> cls, char delimiter) throws IOException {
        CsvMapper csvMapper = new CsvMapper();
        CsvSchema csvSchema = CsvSchema.emptySchema().withHeader().withColumnSeparator(delimiter);
        MappingIterator<I> orderLines = csvMapper.readerFor(cls).with(csvSchema).readValues(data);
        return orderLines.readAll();
    }

    /**
     * 读取csv文件
     *
     * @param csvFilePath 文件地址
     * @param cls         读取转化的对象
     * @param hasHeader   是否存在header
     * @param delimiter   分隔符.默认【,】
     * @param <I>
     * @return
     */
    public static <I> List<I> readCsv(final String csvFilePath, Class<I> cls, boolean hasHeader, char delimiter) {
        if (!isFileExist(csvFilePath)) {
            throw new RuntimeException(String.format("解析用的csv：\u0020[%s] 文件不存在。", csvFilePath));
        }
        if (StrUtil.isEmpty(delimiter)) {
            delimiter = ',';
        }
        try {
            File csvFile = new File(csvFilePath);
            CsvMapper csvMapper = new CsvMapper();
            CsvSchema csvSchema = csvMapper.typedSchemaFor(cls).withStrictHeaders(hasHeader).withColumnSeparator(delimiter).withComments();
            MappingIterator<I> recordIterator = csvMapper.readerWithTypedSchemaFor(cls).with(csvSchema).readValues(csvFile);
            return recordIterator.readAll();
        } catch (IOException e) {
            throw new RuntimeException("[" + csvFilePath + "] 文件解析失败。", e);
        }
    }

    /**
     * 判断文件是否存在
     *
     * @param filePath
     * @return
     */
    public static boolean isFileExist(String filePath) {
        if (StrUtil.isEmpty(filePath)) {
            return false;
        }

        File file = new File(filePath);
        return file.exists() && !file.isDirectory();
    }

    /**
     * 分段读取大文件
     *
     * @param path         文件路径
     * @param readFileBean 分段每次读取的bean 初始值需要设置每次读取的行数
     * @param <T>          读取结果类型bean
     * @return
     */
    public <T> ReadFileBean<T> readBigFile(String path, ReadFileBean<T> readFileBean, Class<T> readCls) {
        File file = new File(path);
        judeFileExists(file);
        try {
            RandomAccessFile randomAccessFile = new RandomAccessFile(file, "r");
            randomAccessFile.seek(readFileBean.getPointer());
            readFileBean.setFileLength(randomAccessFile.length());
            List<String> datas = new ArrayList<>();
            for (int i = 0; i < readFileBean.getReadRowNum(); i++) {
                String s = randomAccessFile.readLine();
                datas.add(charsetDecode(s, StandardCharsets.UTF_8));
                readFileBean.setPointer(randomAccessFile.getFilePointer());
            }
            randomAccessFile.close();
            if (path.contains(TYPE_CSV)) {
                //csv文件处理
                datas.add(0, getCsvHeader(readCls));
                List<T> dataList = readCsv(String.join("\n", datas), readCls, ',');
                readFileBean.setDatas(dataList);
            }
        } catch (FileNotFoundException e) {
            logger.severe("file exists." + e.getMessage());
        } catch (IOException e) {
            logger.severe("read file error." + e.getMessage());
        }
        return readFileBean;
    }

    /**
     * 获取csv的header,使用注解{@link CsvHeader}
     *
     * @param cls
     * @param <T>
     * @return
     */
    public static <T> String getCsvHeader(Class<T> cls) {
        CsvHeader annotation = cls.getAnnotation(CsvHeader.class);
        Assert.notNull(annotation, cls.toString() + "类上需要添加注解@CsvHeader，并指定header。");
        String header = annotation.header();
        Assert.notNull(header, cls.toString() + "类上需要添加注解@CsvHeader，并指定header。");
        return header;
    }

    /**
     * 把字符串data按照指定编码解码
     *
     * @param data    解码字符串
     * @param charset 字符编码
     * @return
     */
    public static String charsetDecode(String data, Charset charset) {
        char[] chars = data.toCharArray();
        byte[] result = new byte[chars.length];
        for (int i = 0; i < chars.length; i++) {
            result[i] = (byte) chars[i];
        }
        return new String(result, charset);
    }

    /**
     * 提供读取IO流方法合集
     * <p>读取速度快慢顺序：</p>
     * 1. {@link #read4Buffer(InputStream, Charset)} <br>
     * 2. {@link #read4IOUtilsCopy(InputStream, Charset)} <br>
     * 3. {@link #read4ByteStreams(InputStream, Charset)} <br>
     * 4. {@link #read4Byte(InputStream, Charset)} <br>
     * 5. {@link #read4StringBuilder(InputStream, Charset)} <br>
     * 6. {@link #read4BufferedReaderParallel(InputStream, Charset, String)}<br>
     * 7. {@link #read4BufferedReader(InputStream, Charset, String)}<br>
     * 8. {@link #read4ScannerA(InputStream)}<br>
     * 9. {@link #read4BufferIO(InputStream, Charset)}<br>
     * 10. {@link #read4IOUtils(InputStream, Charset)}<br>
     * 11. {@link #read4ScannerZ(InputStream)}<br>
     * 12. {@link #read4CharStreams(InputStream, Charset)}<br>
     */
    public static class FileInput {
        /**
         * 读取IO流内容：byte方式
         *
         * @param inputStream
         * @param charset     编码：默认 {@link Charset#defaultCharset()}
         * @return
         * @throws IOException
         */
        public static String read4Byte(InputStream inputStream, Charset charset) throws IOException {
            byte[] bytes = new byte[inputStream.available()];
            inputStream.read(bytes);
            return new String(bytes, charset == null ? Charset.defaultCharset() : charset);
        }

        /**
         * 读取IO流内容：BufferedReader方式
         *
         * @param inputStream
         * @param charset       编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @param lineSeparator 换行方式：默认跟随系统 {@link System#lineSeparator()}
         * @return
         */
        public static String read4BufferedReader(InputStream inputStream, Charset charset, String lineSeparator) {
            return new BufferedReader(
                    new InputStreamReader(inputStream, charset == null ? Charset.defaultCharset() : charset)
            ).lines().collect(Collectors.joining(lineSeparator == null ? System.lineSeparator() : lineSeparator));
        }

        /**
         * 读取IO流内容：BufferedReader 并行方式
         *
         * @param inputStream
         * @param charset       编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @param lineSeparator 换行方式：默认跟随系统 {@link System#lineSeparator()}
         * @return
         */
        public static String read4BufferedReaderParallel(InputStream inputStream, Charset charset, String lineSeparator) {
            return new BufferedReader(
                    new InputStreamReader(inputStream, charset == null ? Charset.defaultCharset() : charset)
            ).lines().parallel()
                    .collect(Collectors.joining(lineSeparator == null ? System.lineSeparator() : lineSeparator));
        }

        /**
         * 读取IO流内容：Scanner A方式
         *
         * @param inputStream
         * @return
         */
        public static String read4ScannerA(InputStream inputStream) {
            Scanner s = new Scanner(inputStream).useDelimiter("\\A");
            String str = s.hasNext() ? s.next() : "";
            return str;
        }

        /**
         * 读取IO流内容：Scanner Z方式
         *
         * @param inputStream
         * @return
         */
        public static String read4ScannerZ(InputStream inputStream) {
            return new Scanner(inputStream).useDelimiter("\\Z").next();
        }

        /**
         * 读取IO流内容：StringBuilder方式
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         */
        public static String read4StringBuilder(InputStream inputStream, Charset charset) throws IOException {
            StringBuilder sb = new StringBuilder();
            String line;
            BufferedReader br = new BufferedReader(new InputStreamReader(inputStream, charset == null ? Charset.defaultCharset() : charset));
            while ((line = br.readLine()) != null) {
                sb.append(line);
            }
            return sb.toString();
        }

        /**
         * 读取IO流内容：ByteArrayOutputStream方式
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         */
        public static String read4Buffer(InputStream inputStream, Charset charset) throws IOException {
            ByteArrayOutputStream result = new ByteArrayOutputStream();
            byte[] buffer = new byte[1024];
            int length;
            while ((length = inputStream.read(buffer)) != -1) {
                result.write(buffer, 0, length);
            }
            return result.toString(charset == null ? Charset.defaultCharset().name() : charset.name());
        }

        /**
         * 读取IO流内容：BufferedInputStream+ByteArrayOutputStream方式
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         */
        public static String read4BufferIO(InputStream inputStream, Charset charset) throws IOException {
            BufferedInputStream bis = new BufferedInputStream(inputStream);
            ByteArrayOutputStream buf = new ByteArrayOutputStream();
            int result = bis.read();
            while (result != -1) {
                buf.write((byte) result);
                result = bis.read();
            }
            return buf.toString(charset == null ? Charset.defaultCharset().name() : charset.name());
        }

        /**
         * 读取IO流内容： 依赖于commons-io:commons-io {@link IOUtils#copy(Reader, OutputStream, Charset)}
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         * @throws IOException
         */
        public static String read4IOUtilsCopy(InputStream inputStream, Charset charset) throws IOException {
            StringWriter writer = new StringWriter();
            IOUtils.copy(inputStream, writer, charset == null ? Charset.defaultCharset() : charset);
            return writer.toString();
        }

        /**
         * 读取IO流内容： 依赖于commons-io:commons-io {@link IOUtils#toString(InputStream, Charset)}
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         * @throws IOException
         */
        public static String read4IOUtils(InputStream inputStream, Charset charset) throws IOException {
            return IOUtils.toString(inputStream, charset == null ? Charset.defaultCharset() : charset);
        }

        /**
         * 读取IO流内容： 依赖于com.google.guava:guava {@link CharStreams#toString(Readable)}
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         * @throws IOException
         */
        public static String read4CharStreams(InputStream inputStream, Charset charset) throws IOException {
            return CharStreams.toString(new InputStreamReader(inputStream, charset == null ? Charset.defaultCharset() : charset));
        }

        /**
         * 读取IO流内容： 依赖于com.google.guava:guava {@link ByteStreams#toByteArray(InputStream)}
         *
         * @param inputStream
         * @param charset     编码：默认跟随系统 {@link Charset#defaultCharset()}
         * @return
         * @throws IOException
         */
        public static String read4ByteStreams(InputStream inputStream, Charset charset) throws IOException {
            return new String(ByteStreams.toByteArray(inputStream), charset == null ? Charset.defaultCharset() : charset);
        }
    }


    /*public static void main(String[] args) {
        long size = FileUtil.size(new File("E:\\OS\\deepin-15.6-amd64\\DeepinCloudPrintServerInstaller_1.0.0.1.exe"));
        System.out.println(size);
        System.out.println(1024 * 1024 * 5);
        if (size > 1024 * 1024 * 5) {
            System.out.println("文件最大5M");
            return;
        }

        long s1 = fileSize(new File("E:\\OS\\cn_windows_10_consumer_editions_version_1803_updated_march_2018_x64_dvd_12063766.iso"));
        System.out.println(s1);
        long s2 = fileSize4Stream(new File("E:\\OS\\cn_windows_10_consumer_editions_version_1803_updated_march_2018_x64_dvd_12063766.iso"));
        System.out.println(s2);

        String s1 = base64(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        System.out.println(s1);

        String s = sha1(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        String s2 = sha1ByBigFile(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        System.out.println(s);
        System.out.println(s2);


        String md5 = md5(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        String md52 = md5ByBigFile(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        System.out.println(md5);
        System.out.println(md52);


        String crc32 = crc32(new File("C:\\Users\\Administrator\\Desktop\\a.html"));
        System.out.println(crc32);
    }*/
}
